import unrealsdk

from ..enums import CurrencyType


class Player:
	'Represents a player object'

	def __init__(self, player):
		'''
			:param unrealsdk.UObject player: Player object to work with
		'''
		self._player = player
		self._player.bHasSeenGoldenKeyMessageThisSession = True


	@classmethod
	def new_from_id(cls, pid) -> 'Mods.ModUtils.Player':
		'''
			Get a local player by ID

			:param int pid: ID of the player to get
		'''
		return cls(unrealsdk.GetEngine().GamePlayers[pid])


	@classmethod
	def default(cls) -> 'Mods.ModUtils.Player':
		'Get the default player (ID:0)'
		return cls.new_from_id(0)


	## player objects
	@property
	def controller(self) -> 'unrealsdk.UObject[WillowGame.WillowController]':
		'A ``WillowController`` associated with the player. Alias for ``LocalPlayer.Actor``.'
		return self._player.Actor


	@property
	def currency(self) -> 'Mods.ModUtils.Currency':
		'A ``Currency`` class associated with the player'
		return Currency(self._player)


	@property
	def pawn(self) -> 'unrealsdk.UObject[WillowGame.WillowPawn]':
		'A ``WillowPawn`` associated with the player. Alias for ``LocalPlayer.Actor.MyWillowPawn``.'
		data = self.controller.MyWillowPawn

		if not data:
			raise AttributeError('No Pawn object. Are you on the main menu?')

		return data


	## useful player properties
	@property
	def character_name(self) -> str:
		'Current character\'s name'
		return self.controller.GetPlayerUINamePreference()


	@property
	def classmod_name(self) -> str:
		'Name of the currently equipted class mod'
		return self.controller.PlayerReplicationInfo.GetClassModName()


	@property
	def level(self) -> int:
		'Level of the current character. Can be modified.'
		self.pawn.GetExpLevel()


	@level.setter
	def level(self, value):
		## need to adjust the max level per game
		# if 1 > level > 80:
		# 	raise ValueError('Level must be anywhere from 1 to 80')

		self.pawn.SetExpLevel(value)


	@property
	def shield_amount(self) -> float:
		'Current shield amount. Can be modified.'
		return self.pawn.GetShieldStrength()


	@shield_amount.setter
	def shield_amount(self, value):
		max = self.shield_amount_max

		if 0 > value > max:
			raise ValueError(f'Shield amount must be between 0 and {max}')

		self.pawn.SetShieldStrength(value)


	@property
	def shield_amount_max(self) -> float:
		'Max amount of the currently equipted shield'
		return self.pawn.GetMaxShieldStrength()


	@property
	def user_name(self) -> str:
		'Your account name. Usually your Steam display name.'
		return self._player.GetNickname()


	def get_net_speed(self) -> 'tuple[int,int]':
		'Currently configured speeds for ``internet`` and ``lan``. Format is ```(internet, lan)```'
		player = self.controller.Player

		return player.ConfiguredInternetSpeed, player.ConfiguredLanSpeed


	def golden_key_add(self, value):
		'''
			Add x amount of golden keys

			:param int value: Number of keys to add
		'''
		self.controller.AddGoldenKeysFromSource(0, value)


	def golden_key_remove(self, value):
		'''
			Remove x amount of golden keys (WARNING: untested)

			:params int value: Number of keys to remove
		'''
		controller = self.controller

		for _ in range(0, value - 1):
			controller.SpendGoldenKey()


	def console_command(self, *cmd):
		'''
			Run a console command. All argument values will be converted to ``str`` objects and
			joined with a space before being sent.

			:param tuple(typing.Any) cmd: Parts of the console command
		'''
		command = ' '.join(str(c) for c in cmd)
		self.controller.ConsoleCommand(command)


	def refill_life(self):
		'Set the current character\'s life to max'
		self.pawn.FullyReplenishLife()


	def refill_shields(self):
		'Set the current character\'s shield to max'
		self.pawn.FullyReplenishShields()


	def remove_status_effects(self):
		'Remove all active status effects'
		self.pawn.RemoveAllStatusEffects()


	def set_net_speed(self, internet=None, lan=None):
		'''
			Set max net speeds

			:param int internet: Max speed to set for internet connections
			:param int lan: Max speed to set for lan connections
		'''
		player = self.controller.Player

		if lan == internet == None:
			raise ValueError('Must provide a value for `lan` and/or `internet`')

		if lan != None:
			player.ConfiguredLanSpeed = lan

		if internet != None:
			player.ConfiguredInternetSpeed = internet



class Currency:
	'Manage a character\'s currency. The types can be accessed via dict items or object attributes.'

	def __init__(self, player):
		'''
			:param Player player: Player object to work with
		'''
		self._player = player
		self._repinfo = player.controller.PlayerReplicationInfo


	def __getattr__(self, key):
		try:
			get = object.__getattribute__(self, 'get')
			return get(key)

		except KeyError:
			pass

		object.__getattribute__(self, key)


	def __setattr__(self, key, value):
		try:
			set = object.__getattribute__(self, 'set')
			set(key, value)

		except KeyError:
			pass

		object.__setattr__(self, key, value)


	def __getitem__(self, key):
		return self.get(key)


	def __setitem__(self, key, value):
		self.set(key, value)


	@classmethod
	def new_from_player_id(cls, pid) -> 'Currency':
		'''
			Create a new ``Currency`` object for a player with the specified ID

			:param int pid: ID of the player
		'''
		return cls(Player.new_from_id(pid))


	@classmethod
	def default(cls) -> 'Currency':
		'Create a new ``Currency`` object for the default player (ID: 0)'
		return cls(Player.default())


	def add(self, type, value):
		'''
			Add x amount of the specified currency

			:param typing.Union[CurrencyType,str] type: Type of currency to work with
			:param int value: Amount to add
		'''
		if not isinstance(type, CurrencyType):
			type = CurrencyType.parse(type)

		self._repinfo.AddCurrencyOnHand(type.value)


	def get(self, type):
		'''
			Get the current amount of the specified currency

			:param typing.Union[CurrencyType,str] type: Type of currency to get
		'''
		if not isinstance(type, CurrencyType):
			type = CurrencyType.parse(type)

		self._repinfo.GetCurrencyOnHand(type.value)


	def set(self, type, value):
		'''
			Set the amount of the specified currency

			:param typing.Union[CurrencyType,str] type: Type of currency to work with
			:param int value: Amount to set the currency to
		'''
		if not isinstance(type, CurrencyType):
			type = CurrencyType.parse(type)

		self._repinfo.SetCurrencyOnHand(type.value, value)


	def subtract(self, type, value):
		'''
			Subtraft x amount of the specified currency

			:param typing.Union[CurrencyType,str] type: Type of currency to work with
			:param int value: Amount to subtract
		'''
		if not isinstance(type, CurrencyType):
			type = CurrencyType.parse(type)

		self._repinfo.AddCurrencyOnHand(type.value, value * -1)


	def to_dict(self) -> 'dict[str,int]':
		'Get all currencies and their amounts'
		values = self._repinfo.GetAllCurrencyOnHand()
		data = {}

		for idx, type in enumerate(CurrencyType):
			data[type.name] = values[idx]

		return data


class SkillTree:
	'View and/or manage a character\'s skill tree (WIP)'

	def __init__(self, player):
		'''
			:param Player player: Player object to work with
		'''
		self._player = player
		self._tree = player.controller.PlayerSkillTree


	@classmethod
	def new_from_player_id(cls, pid) -> 'SkillTree':
		'''
			Create a new ``SkillTree`` object for a player with the specified ID

			:param int pid: ID of the player
		'''
		return cls(Player.new_from_id(pid))


	@classmethod
	def default(cls) -> 'SkillTree':
		'Create a new ``SkillTree`` object for the default player (ID: 0)'
		return cls(Player.default())
