from Mods.ModMenu import (
	KeybindManager as KBM,
	Options
)

from Mods.ModMenu.ModObjects import EnabledSaveType, Game, ModPriorities, ModTypes, RegisterMod, SDKMod
from Mods.ModMenu.Options import Boolean, Slider, Spinner
from dataclasses import dataclass, field
from unrealsdk import Log

from .misc import clsname, combine_flags, is_in_game, manifest_to_flag
from .objects.player import Player


MOD_VALUES = {
	'Name': 'name',
	'Author': 'author',
	'Description': 'description',
	'Version': 'version',
	'SupportedGames': 'supported_games',
	'Types': 'types',
	'Priority': 'priority',
	'SaveEnabledState': 'save_enabled_state',
	'Status': 'status',
	'SettingsInput': 'settings_input',
	'Options': 'options',
	'Keybinds': 'keybinds'
}

SAVE_STATES = {
	'menu': EnabledSaveType.LoadOnMainMenu,
	'settings': EnabledSaveType.LoadWithSettings,
	'never': EnabledSaveType.NotSaved
}


class SdkMod(SDKMod):
	'''
		Convenience class for ``ModMenu.SDKMod``. Mod options can be accessed like dict items.
	'''

	def __new__(cls, *args, **kwargs):
		instance = SDKMod.__new__(cls, *args, **kwargs)
		instance.Keybinds = KeybindList.new_init(cls.Keybinds)
		instance.Options = OptionsList.new_init(cls.Options)
		instance.SettingsInputs = SettingsInputDict.new_init(cls.SettingsInputs)

		return instance


	def __getattr__(self, key):
		if key == 'player':
			return Player.default()

		elif key == 'world':
			return World.default()

		return SDKMod.__getattribute__(self, key)


	def __getitem__(self, key):
		return self.Options.get(key).CurrentValue


	def __setitem__(self, key, value):
		self.Options.set(key, value)


	def __delitem__(self, key):
		self.Options.set(key)


	def get_enable_state(self) -> EnabledSaveType:
		'Get the stage at which the mod gets auto-enabled'
		return self.SaveEnabledState


	def get_games(self) -> Game:
		'Get the flag of supported games'
		return self.SupportedGames


	def get_keybind(self, name) -> KBM.Keybind:
		'''
			Get a keybind

			:param str name: Name of the keybind to get
		'''
		return self.Keybinds.get(name)


	def get_option(self, key) -> 'typing.Any':
		'''
			Get a mod option

			:param str key: Name of the option to get
		'''
		return self.Options.get(key).CurrentValue


	def get_priority(self) -> 'typing.Union[ModPriorities,int]':
		'Get the init priority'
		return self.Priority


	def get_settings_input(self, key) -> str:
		'''
			Get the label for a settings input key

			:param typing.Union[KeyboardInput,str] key: Input key
		'''
		return self.SettingsInput[key]


	def get_types(self) -> ModTypes:
		'Get the mod categories'
		return self.Types


	def log(self, *data):
		'Output some text to the in-game console. It will be prefixed with "[{modname}]".'
		message = ' '.join(str(msg) for msg in data)
		Log(f'[{self.Name}] {message}')


	def register(self):
		'Register the mod with the mod manager'
		if is_in_game():
			RegisterMod(self)


	def set_enable_state(self, state):
		'''
			Set the stage at which the mod will be auto-enabled

			:param EnabledSaveType state: Stage at which the mod will be auto-enabled
		'''
		if isinstance(state, str):
			state = SAVE_STATES.get(state)

			if not state:
				state = EnabledSaveType[str]

		elif not isinstance(state, EnabledSaveType):
			raise TypeError(f'Mod enable state must be a str or EnabledSaveType, not {clsname(priority)}')

		self.SaveEnabledState = state


	def set_games(self, *values):
		'''
			Set the games this mod supports

			:param tuple[typing.Union[Game,str]] values: Supported games
		'''
		items = []

		for value in values:
			if isinstance(value, str):
				value = Game[value]

			elif not isinstance(value, Game):
				raise TypeError(f'Game type must be a str or Game, not {(clsname(value))}')

			items.append(value)

		self.SupportedGames = combine_flags(*items)


	def set_keybind(self, name, key, rebindable=True, hidden=False, handler=None) -> KBM.Keybind:
		'''
			Create a new in-game keybind

			:param str name: Display name of the keybind.
			:param typing.Any[KeyboardInput,str] key: Key to assign to the keybind.
			:param bool rebindable: Whether or not the key can be re-bound.
			:param bool hidden: Whether or not to display the keybind in the options menu.
			:param typing.Callable handler: Function to run when the key is pressed. Leave empty to use the default handler.
		'''
		return self.Keybinds.new(name, key, rebindable, hidden, handler)


	def set_option(self, key, value=None):
		'''
			Set a value for a mod option. Leaving ``value`` as ``None`` sets the option to default.

			:param str key: Name of the option to change
			:param typing.Any value: New value for the option
		'''
		self.Options.set(key, value)


	def set_priority(self, priority):
		'''
			Set the init priority of the mod

			:param typing.Union[ModPriorities,int] priority: Init priority
		'''
		if not instance(priority, (int, ModPriorities)):
			raise TypeError(f'Priority must be an int or ModPriorities, not {clsname(priority)}')

		self.Priority = priority


	def set_settings_input(self, key, label):
		'''
			Create a new input key for the mod on the Mods page

			:param typing.Union[KeyboardInput,str] key: Key to be used to activate the action
			:param str label: Name to display for the key
		'''
		self.SettingsInput.set(key, label)


	def set_types(self, *types):
		'''
			Set the mod's categories

			:param tuple[typing.Union[ModTypes,str]] types: Mod categories
		'''
		items = []

		for value in types:
			if isinstance(value, str):
				value = ModTypes[value]

			elif not isinstance(value, ModTypes):
				raise TypeError(f'Mod type must be a str or ModTypes, not {clsname(value)}')

			items.append(value)

		self.Types = combine_flags(*items)


	## New handler methods for built-in events
	def handle_disable(self):
		'Called when the mod gets disabled. Override this in a subclass.'


	def handle_enable(self):
		'Called when the mod gets enabled. Override this in a subclass.'


	def handle_game_input(self, bind, event):
		'''
			Called when a mod keybind is pressed. Override this in a subclass.

			:param ModMenu.Keybind bind: Keybind associated with the input
			:param ModMenu.InputEvent event: Type of input
		'''


	@staticmethod
	def handle_network_deserialize(data):
		'''
			Called when data is recieved from a network function. Calls json.loads by default.
			Override this in a subclass.

			:param str data: JSON data recieved from the network
		'''


	@staticmethod
	def handle_network_serialize(args):
		'''
			Called when data is to be sent over the network. Calls json.dumps by default.
			Override this in a subclass.

			:param dict data: Data to be sent over the network
		'''


	def handle_option_change(self, option, value):
		'''
			Called when a mod option has been changed. Override this in a subclass.

			:param Mods.ModMenu.Options.Value option: Option that recieved the new setting
			:param typing.Any value: New value to be set to the option
		'''


	def handle_settings_input(self, action):
		'''
			Called when a keybind for the mod menu has been pressed. Override this in a subclass.

			:param str action: Name of the action associated with the input
		'''


	## Overriding functions and events to add handler functions for easier sub-classing
	@classmethod
	def NetworkDeserialize(cls, data):
		args = cls.handle_network_deserialize(data)

		if not args or isinstance(args, str):
			return SDKMod.NetworkDeserialize(args)

		return args


	@classmethod
	def NetworkSerialise(cls, args):
		data = cls.handle_network_serialize(args)

		if not data or isinstance(data, dict):
			return SDKMod.NetworkSerialise(data)

		return data


	def Disable(self):
		SDKMod.Disable(self)
		self.handle_disable()


	def Enable(self):
		SDKMod.Enable(self)
		self.handle_enable()


	def GameInputPressed(self, bind, event):
		self.handle_game_input(bind, event)


	def ModOptionChanged(self, option, value):
		self.handle_option_change(option, value)


	def SettingsInputPressed(self, action):
		SDKMod.SettingsInputPressed(self, action)
		self.handle_settings_input(action)


class KeybindList(list):
	def __init__(self, *keybinds):
		list.__init__(self)

		for keybind in keybinds:
			self.add(keybind)


	@classmethod
	def new_init(cls, keybinds):
		if not keybinds:
			keybinds = []

		return cls(*keybinds)


	def add(self, keybind):
		'''
			Add an existing Keybind object to the keybind list
		'''
		if not isinstance(keybind, KBM.Keybind):
			raise TypeError('Keybind is not a ModMenu.Keybind object')

		if self.get(keybind.Name):
			raise ValueError(f'Keybind with id already exists: {keybind.Name}')

		self.append(keybind)
		return keybind


	def get(self, name):
		'''
			Get an in-game keybind
		'''
		for key in self:
			if name in {keybind.Name, keybind.Name.replace(' ', '')}:
				return key


	def new(self, name, key, rebindable=True, hidden=False, handler=None):
		'''
			Create a new in-game keybind for an action
		'''
		keybind = KBM.Keybind(
			Name = name,
			Key = key,
			IsRebindable = rebindable,
			IsHidden = hidden,
			OnPress = handler
		)

		return self.add(keybind)


class OptionsList(list):
	def __init__(self, *options):
		list.__init__(self)

		for opt in options:
			self.add(opt)


	@classmethod
	def new_init(cls, options):
		if not options:
			options = []

		return cls(*options)


	def add(self, opt, id=None):
		if not isinstance(opt, Options.Value):
			raise TypeError('Option must be a ModUtils.BaseValue and ModMenu.Options.Value object')

		opt.id = id or opt.Caption.replace(' ', '')
		self.append(opt)
		return opt


	def get(self, id):
		for opt in self:
			if id in {opt.id, opt.Caption}:
				return opt


	def new_boolean(self, name, description, default=False, true_value='On', false_value='Off', id=None):
		return self.add(Options.Boolean(
			Caption = name,
			Description = description,
			StartingValue = default,
			Choices = (false_value, true_value)
		), id)


	def new_slider(self, name, description, default=1, min=0, max=100, increment=1, id=None):
		return self.add(Options.Slider(
			Caption = name,
			Description = description,
			StartingValue = default,
			MinValue = min,
			MaxValue = max,
			Increment = increment
		), id)


	def new_spinner(self, name, descrption, choices: list, default=None, id=None):
		return self.add(Options.Spinner(
			Caption = name,
			Description = description,
			StartingValue = default or choices[0]
		), id)


	def set(self, id, value=None):
		opt = self.get(id)
		opt.CurrentValue = value if value != None else opt.StartingValue


	def to_dict(self):
		return {opt.id: opt.CurrentValue for opt in self}


class SettingsInputDict(dict):
	def __init__(self, **binds):
		dict.__init__(self, {'Enter': 'Enable'})

		for keybind, name in binds.items():
			self.set(name, keybind)


	@classmethod
	def new_init(cls, binds):
		if not binds:
			binds = {}

		return cls(**binds)


	def get_by_name(self, name):
		for keybind, keybind_name in self.items():
			if keybind_name == name:
				return keybind

		raise KeyError(name)


	def set(self, key, name):
		self[key] = name


# class SdkManifestMod(SDKMod):
# 	def __init__(self, dirname):
# 		with Path(__file__).parent.joinpath(f'../{dirname}/manifest.json').open('r') as manifest:
# 			self.manifest = json.load(manifest)
# 
# 
# 	def __getattr__(self, key):
# 		if key in MOD_VALUES:
# 			key = MOD_VALUES[key]
# 
# 		return object.__getattribute__(self, key)
# 
# 
# 	def __setattr__(self, key, value):
# 		if key in MOD_VALUES:
# 			key = MOD_VALUES[key]
# 
# 		object.__setattr__(self, key, value)
# 
# 
# 	def __delattr__(self, key):
# 		if key in MOD_VALUES:
# 			key = MOD_VALUES[key]
# 
# 		return object.__delattr__(self, key)
# 
# 
# 	@property
# 	def Author(self):
# 		return self.manifest['authors']
# 
# 
# 	@property
# 	def Description(self):
# 		return self.manifest.get('tagline', self.manifest['description'])
# 
# 
# 	@property
# 	def Name(self):
# 		return self.manifest['name']
# 
# 
# 	@property
# 	def Priority(self):
# 		value = self.manifest['modinfo']['priority']
# 
# 		if isinstance(value, str):
# 			return ModPriorities[value.title()]
# 
# 		return value
# 
# 
# 	@property
# 	def SaveEnabledState(self):
# 		return EnabledSaveType[self.manifest['modinfo']['save_state']]
# 
# 
# 	@property
# 	def SupportedGames(self):
# 		return manifest_to_flag(Game, self.manifest['supports'])
# 
# 
# 	@property
# 	def Types(self):
# 		return manifest_to_flag(ModTypes, self.manifest['types'])
# 
# 
# 	@property
# 	def Version(self):
# 		return self.manifest['latest']
